import pytest


def test_node_initialization(node1):
    assert node1.tag == "Test One"
    assert node1.id == "identifier 1"
    assert node1.expanded
    assert node1.parent is None
    assert isinstance(node1.children, list)
    assert node1.children == []
    assert node1.data is None


def test_node_repr(node1):
    assert repr(node1) == "Node(tag='Test One', id='identifier 1', data=None)"


def test_node_set_tag(node1):
    node1.tag = "Test 1"
    assert node1.tag == "Test 1"


def test_node_set_identifier(node1):
    node1.id = "ID1"
    assert node1.id == "ID1"

    with pytest.raises(ValueError):
        node1.id = None

    node1._set_id(None)
    assert node1.id.version == 1


def test_node_add_child(node1):
    node1.add_child('id 2')
    assert node1.children == ['id 2']


def test_node_parent(node2):
    node2.parent = "id 1"
    assert node2.parent == 'id 1'


def test_node_is_leaf(node1, node2):
    node1.add_child('id 2')
    node2.parent = "id 1"
    assert not node1.is_leaf
    assert node2.is_leaf


def test_node_data(node1):
    class Flower(object):
        def __init__(self, color):
            self.color = color

        def __str__(self):
            return "%s" % self.color

    node1.data = Flower("red")
    assert node1.data.color == "red"



import pytest

from ttree import Tree, Node


@pytest.fixture
def node1():
    return Node("Test One", "identifier 1")


@pytest.fixture
def node2():
    return Node("Test Two", "identifier 2")


@pytest.fixture
def tree():
    result = Tree()
    result.create_node("Hárry", "hárry")
    result.create_node("Jane", "jane", parent="hárry")
    result.create_node("Bill", "bill", parent="hárry")
    result.create_node("Diane", "diane", parent="jane")
    result.create_node("George", "george", parent="bill")
    """
    Hárry
    ├── Bill
    │   └── George
    └── Jane
        └── Diane
    """
    return result


@pytest.fixture
def tree_as_string():
    return """\
Hárry
|-- Jane
|   +-- Diane
+-- Bill
    +-- George
"""

@pytest.fixture
def copytree(tree):
    return Tree(tree, True)



"""
Taxonomy tree data structure implementation
===========================================

`ttree` is a Python module with two primary classes: ``Node`` and
``Tree``. Tree is a self-contained structure with some nodes and
connected by branches. A tree owns merely a root, while a
node (except root) has some children and one parent.
"""
__version__ = '1.0.1'

from .tree import Tree  # noqa
from .node import Node  # noqa



#!/usr/bin/env python
# A file folder scanner contributed by @holger 
#
# You can specify the scanned folder and file pattern by changing root_path
# and pattern variables
#
__author__ = 'holger'

import argparse
from hashlib import blake2b
from pathlib import Path

from ttree import tree

FILECOUNT = 0
DIRCOUNT = 0
DIR_ERRORLIST = []
FILE_ERRORLIST = []


parser = argparse.ArgumentParser(description='Scan the given folder and print '
                                             'its structure in a tree.')
parser.add_argument('path', nargs='?', type=Path,
                    help='An path to be scanned.',
                    default=Path.cwd())
parser.add_argument('pattern', nargs='?', type=str,
                    help='File name pattern to filtered, e.g. *.pdf',
                    default="*")
parser.add_argument('--debug', type=bool, nargs='?', const=True, default=False,
                    help='Debug mode')
parser.add_argument('--profiling', type=str, choices=['timeit', 'cprofile'],
                    default=None, help='Profiling mode')

args = parser.parse_args()
root_path = Path(args.path).resolve(strict=True)

folder_tree = tree.Tree()
root_node = folder_tree.create_node(str(root_path), root_path)  # root node


def get_hash(data):
    data = bytes(str(data), 'UTF-8')
    data_hash = blake2b(data, digest_size=8).hexdigest()

    if args.debug:
        print('++++++ blake2b ++++++')
        print('input: ' + str(data))
        print('hash: ' + data_hash)
        print('+++++++++++++++++++++')
    return data_hash


def get_node_id(depth, path):
    """ get_node_id returns
        - depth contains the current depth of the folder hierarchy
        - folder contains the current directory

        Function returns a string containing the current depth, the folder name
        and unique ID build by hashing the absolute path of the directory.
        All spaces are replaced by '_'

        <depth>_<dirname>+++<hash>
        e.g. 2_Folder_XYZ_1+++<hash>
    """
    return f"{depth}_{path}+++{get_hash(path)}".replace(" ", "_")


def get_parent_id(depth, path):
    # special case for the 'root' of the tree
    # because we don't want a cryptic root-name
    if path == root_path:
        return str(root_path)

    # looking for parent directory
    # e.g. /home/user1/mp3/folder1/parent_folder/current_folder
    # get 'parent_folder'

    parent_id = f"{depth - 1}_{path.parent}+++{get_hash(path.parent)}"
    return parent_id.replace(" ", "_")


def print_node(folder, node_id, parent_id):
    print(f"""
#############################
node created
folder: {folder}
node_id: {node_id}
parent_id: {parent_id}
""")


def crawler():
    global DIRCOUNT
    global FILECOUNT

    start_depth = len(root_path.parts)

    for current_path in sorted(root_path.glob(f'**/{args.pattern}')):
        current_depth = len(current_path.parts) - start_depth
        if args.debug:
            print(f'current: {current_path}')

        node_id = get_node_id(current_depth, current_path)
        parent_id = get_parent_id(current_depth, current_path)

        if args.debug:
            print_node(current_path, node_id, parent_id)

        if parent_id not in folder_tree:
            parent_id = root_node

        # create node
        folder_tree.create_node(current_path.name, node_id, parent=parent_id)

        # +++ DIRECTORIES +++
        if current_path.is_dir():
            if parent_id == str(None):
                DIR_ERRORLIST.append(current_path)
            DIRCOUNT += 1
        elif current_path.is_file():
            # +++ FILES +++
            if parent_id == str(None):
                FILE_ERRORLIST.append(current_path)
            FILECOUNT += 1


profiling_result = None

if args.profiling is None:
    crawler()
elif args.profiling == 'timeit':
    import timeit
    timeit_result = timeit.Timer("crawler()", "from __main__ import crawler")
    profiling_result = f'time: {timeit_result.timeit(number=1)}'
elif args.profiling == 'cprofile':
    import cProfile
    profile = cProfile.Profile()
    profiling_result = profile.run("crawler()")

folder_tree.print()

if DIR_ERRORLIST:
    for item in DIR_ERRORLIST:
        print(item)
else:
    print('no directory errors')

if FILE_ERRORLIST:
    for item in FILE_ERRORLIST:
        print(item)
else:
    print('no file errors')

print(f'Count of files: {FILECOUNT}')
print(f'Count of folders: {DIRCOUNT}')
print(f'Count of tree nodes: {len(folder_tree)}')

if profiling_result is not None:
    print('\nProfiling results:')
    if isinstance(profiling_result, str):
        print(profiling_result)
    elif isinstance(profiling_result, cProfile.Profile):
        profiling_result.print_stats('ncalls')



#!/usr/bin/env python
# Example usage of ttree
#
# Author: chenxm
#
__author__ = 'chenxm'

from ttree import Tree


def create_family_tree():
    # Create the family tree
    tree = Tree()
    tree.create_node("Harry", "harry")  # root node
    tree.create_node("Jane", "jane", parent="harry")
    tree.create_node("Bill", "bill", parent="harry")
    tree.create_node("Diane", "diane", parent="jane")
    tree.create_node("Mary", "mary", parent="diane")
    tree.create_node("Mark", "mark", parent="jane")
    return tree


def example(text):
    print(f'\n{" " + text + " ":-^80}\n')


if __name__ == '__main__':
    tree = create_family_tree()

    example("Tree of the whole family")
    tree.print(key=lambda x: x.tag, reverse=True, ascii_mode='em')

    example("All family members in DEPTH mode")
    print(', '.join([tree[node].tag for node in tree.expand_tree()]))

    example("All family members (with identifiers) but Diane's sub-family")
    tree.print(id_hidden=False, filtering=lambda x: x.id != 'diane')

    example("Let me introduce Diane family only")
    tree.subtree('diane').print()

    example("Children of Diane")
    for child in tree.is_branch('diane'):
        print(tree[child].tag)

    example("New members join Bill's family")
    new_tree = Tree()
    new_tree.create_node("n1", 1)  # root node
    new_tree.create_node("n2", 2, parent=1)
    new_tree.create_node("n3", 3, parent=1)
    tree.paste('bill', new_tree)
    tree.print()

    example("They leave after a while")
    tree.remove_node(1)
    tree.print()

    example("Now Mary moves to live with grandfather Harry")
    tree.move_node('mary', 'harry')
    tree.print()

    example("A big family for Mark to send message to the oldest Harry")
    print(', '.join([tree[node].tag for node in tree.rsearch('mark')]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

from ttree import Tree
from ttree.utils import export_to_dot


def read_generated_output(filename):
    with open(filename, 'r', encoding='utf-8') as fp:
        return fp.read()


def test_export_to_dot(tree):
    export_to_dot(tree, 'tree.dot')
    expected = """\
digraph tree {
\t"hárry" [label="Hárry", shape=circle]
\t"bill" [label="Bill", shape=circle]
\t"jane" [label="Jane", shape=circle]
\t"george" [label="George", shape=circle]
\t"diane" [label="Diane", shape=circle]

\t"hárry" -> "jane"
\t"hárry" -> "bill"
\t"bill" -> "george"
\t"jane" -> "diane"
}"""

    assert os.path.isfile('tree.dot')
    generated = read_generated_output('tree.dot')

    assert generated == expected
    os.remove('tree.dot')


def test_export_to_dot_empty_tree():
    empty_tree = Tree()
    export_to_dot(empty_tree, 'tree.dot')

    expected = """\
digraph tree {

}"""
    assert os.path.isfile('tree.dot')
    generated = read_generated_output('tree.dot')

    assert expected == generated
    os.remove('tree.dot')


def test_unicode_filename():
    tree = Tree()
    tree.create_node('Node 1', 'node_1')
    export_to_dot(tree, 'ŕʩϢ.dot')

    expected = """\
digraph tree {
\t"node_1" [label="Node 1", shape=circle]

}"""
    assert os.path.isfile('ŕʩϢ.dot')
    generated = read_generated_output('ŕʩϢ.dot')
    assert expected == generated
    os.remove('ŕʩϢ.dot')


def test_export_with_minus_in_filename():
    tree = Tree()
    tree.create_node('Example Node', 'example-node')
    expected = """\
digraph tree {
\t"example-node" [label="Example Node", shape=circle]

}"""

    export_to_dot(tree, 'id_with_minus.dot')
    assert os.path.isfile('id_with_minus.dot')
    generated = read_generated_output('id_with_minus.dot')
    assert expected == generated
    os.remove('id_with_minus.dot')



class NodeNotFound(Exception):
    """Exception raises if a node's identifier is unknown"""
    pass


class NodePropertyNotFound(Exception):
    """Exception raises if a node's data property is not specified"""
    pass


class MultipleRoots(Exception):
    """Exception raises if more than one root exists in a tree."""
    pass


class DuplicatedNode(Exception):
    """Exception raises if an identifier already exists in a tree."""
    pass


class LinkPastRootNode(Exception):
    """
    Exception raises in Tree.link_past_node() if one attempts
    to "link past" the root node of a tree.
    """
    pass


class InvalidLevelNumber(Exception):
    pass


class LoopError(Exception):
    """
    Exception raises if trying to move node B to node A's position
    while A is B's ancestor.
    """
    pass



__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}.{{patch}}',
}

FILES = [
    {
        'path': 'ttree/__init__.py',
        'serializer': "__version__ = '{{ GLOBALS.serializer }}'"
    },
]

VERSION = ['major', 'minor', 'patch']

VCS = {
    'name': 'git',
    'commit_message': "Version updated from {{ current_version }} to {{ new_version }}",
}



from enum import Enum


class ASCIIMode(Enum):
    simple = ('|', '|-- ', '+-- ')
    ex = ('│', '├── ', '└── ')
    exr = ('│', '├── ', '╰── ')
    em = ('║', '╠══ ', '╚══ ')
    emv = ('║', '╟── ', '╙── ')
    emh = ('│', '╞══ ', '╘══ ')


class TraversalMode(Enum):
    #: The depth-first search mode for tree.
    DEPTH = 'depth'
    #: The width-first search mode for tree.
    WIDTH = 'width'
    #: `ZIGZAG search <https://en.wikipedia.org/wiki/Tree_(data_structure)>`_
    #: mode for tree.
    ZIGZAG = 'zigzag'



#!/usr/bin/env python

"""
Example of ttree usage to generate recursive tree of directories.
It could be useful to implement Directory Tree data structure

2016 samuelsh
"""

import ttree
import random
from hashlib import blake2b
from string import digits, ascii_letters

MAX_FILES_PER_DIR = 10


def get_random_string(length):
    return ''.join(random.choice(digits + ascii_letters) for _ in range(length))


def build_recursive_tree(tree, base, depth, width):
    """
    Args:
        tree: Tree
        base: Node
        depth: int
        width: int

    Returns:

    """
    while depth >= 0:
        depth -= 1
        for _ in range(width):
            directory = Directory()
            tree.create_node(
                directory.name,
                # node id is md5 hash of it's name
                blake2b(directory.name.encode()).hexdigest(),
                parent=base.id,
                data=directory
            )
        dirs_nodes = tree.children(base.id)
        for dir in dirs_nodes:
            newbase = tree.get(dir.id)
            build_recursive_tree(tree, newbase, depth, width)


class Directory:
    def __init__(self):
        self._name = get_random_string(64)
        # Each directory contains 1000 files
        self._files = [File() for _ in range(MAX_FILES_PER_DIR)]

    @property
    def name(self):
        return self._name

    @property
    def files(self):
        return self._files


class File:
    def __init__(self):
        self._name = get_random_string(64)

    @property
    def name(self):
        return self._name


tree = ttree.Tree()
base = tree.create_node('Root', 'root')
build_recursive_tree(tree, base, 2, 10)

tree.print()



from typing import Hashable, Callable, Union

from ttree.common import ASCIIMode, TraversalMode


def tree_printer_gen(tree, node_id: Hashable, filtering=None, key=None,
                     reverse: bool = False,
                     ascii_mode: ASCIIMode = ASCIIMode.ex,
                     is_last: list = None, level: int = 0):
    """
    Generate elements of existing tree.

    :param ~ttree.Tree tree: Tree instance
    :param node_id: Traversal root node ID
    :param filtering: Filtering callable
    :param key: Sorting key callable
    :param reverse: Reverse mode
    :param ascii_mode: ASCII mode
    :param is_last: Recursion tail flag
    :param level: Current level of tree in recursion
    """
    c_line, c_branch, c_corner = ascii_mode.value

    if is_last is None:
        is_last = []

    node_id = tree.root if node_id is None else node_id
    node = tree[node_id]

    if level == 0:
        yield "", node
    else:
        leading = ''.join(c_line + ' ' * 3 if not x else ' ' * 4
                          for x in is_last[0:-1])
        lasting = c_corner if is_last[-1] else c_branch
        yield leading + lasting, node

    if node.expanded:
        if filtering is not None and not filtering(node):
            return

        children = (tree[n] for n in node.children)

        if filtering is not None:
            children = filter(filtering, children)

        if key is not None:
            children = (n for n in sorted(children, key=key, reverse=reverse))
        elif reverse:
            children = (n for n in reversed(list(children)))

        level += 1
        current_child = next(children, None)
        while children and current_child is not None:
            next_child = next(children, None)
            is_last.append(next_child is None)
            for item in tree_printer_gen(tree, current_child.id,
                                         filtering, key, reverse, ascii_mode,
                                         is_last, level=level):
                yield item
            is_last.pop()
            current_child = next_child


def get_label(node, data_property: str, id_hidden: bool):
    """
    Get label of node.

    :param ~ttree.Node node: Tree node instance
    :param data_property: Data property name
    :param id_hidden: Is ID hidden?
    :return:
    """
    result = getattr(node.data, data_property) \
        if data_property \
        else node.tag

    return result if id_hidden else f'{result}[{node.id}]'


def print_tree(tree, node_id: Hashable = None, id_hidden: bool = True,
               filtering=None, key=None, reverse: bool = False,
               ascii_mode: Union[ASCIIMode, str] = ASCIIMode.ex,
               data_property: str = None, func: Callable = None):
    """
    Another implementation of printing tree using Stack Print tree structure
    in hierarchy style.

    For example:

    .. code-block:: text

        Root
        |___ C01
        |    |___ C11
        |         |___ C111
        |         |___ C112
        |___ C02
        |___ C03
        |    |___ C31

    A more elegant way to achieve this function using Stack
    structure, for constructing the Nodes Stack push and pop nodes
    with additional level info.

    UPDATE: the @key @reverse is present to sort node at each
    level.

    :param ~ttree.Tree tree: Tree instance
    :param node_id: Traversal root node ID
    :param id_hidden: Is ID hidden?
    :param filtering: Filtering callable
    :param key: Sorting key callable
    :param reverse: Reverse mode?
    :param ascii_mode: ASCII mode
    :param data_property: Data property name
    :param func: Printer function callable
    """
    ascii_mode = (
        ascii_mode if isinstance(ascii_mode, ASCIIMode)
        else ASCIIMode[ascii_mode]
    )

    result = (
        f'{pre}{get_label(node, data_property, id_hidden)}'
        for pre, node in tree_printer_gen(tree, node_id, filtering, key,
                                          reverse, ascii_mode)
    )

    if func is not None and callable(func):
        for s in result:
            func(s)
    else:
        return '\n'.join(s for s in result) + '\n'


def export_to_dot(tree, filename: str, shape='circle', graph='digraph'):
    """
    Export the tree in the DOT format of the Graphviz software.

    .. seealso::

        * `Graphviz <http://www.graphviz.org>`_
        * `DOT Language <http://www.graphviz.org/content/dot-language>`_

    :param ~ttree.Tree tree: Tree instance
    :param filename: Export file name
    :param shape: Shape of tree node in DOT
    :param graph: Graph type in DOT
    """
    nodes, connections = [], []

    if tree:
        for node in tree.expand_tree(mode=TraversalMode.WIDTH):
            node_id = tree[node].id
            nodes.append(
                f'"{node_id}" [label="{tree[node].tag}", shape={shape}]'
            )

            connections.extend(
                [f'"{node_id}" -> "{c.id}"' for c in tree.children(node_id)]
            )

    # write nodes and connections to dot format
    with open(filename, 'w', encoding='utf-8') as f:
        f.write(f'{graph} tree {{\n')

        for node in nodes:
            f.write(f'\t{node}\n')

        f.write('\n')

        for child in connections:
            f.write(f'\t{child}\n')

        f.write('}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest

from ttree import Tree, Node
from ttree.exceptions import NodeNotFound, LoopError


def test_tree(tree, copytree):
    assert isinstance(tree, Tree)
    assert isinstance(copytree, Tree)


def test_is_root(tree):
    assert tree['hárry'].is_root
    assert not tree['jane'].is_root


def test_paths_to_leaves(tree):
    paths = tree.paths_to_leaves
    assert len(paths) == 2
    assert ['hárry', 'jane', 'diane'] in paths
    assert ['hárry', 'bill', 'george'] in paths


def test_nodes(tree):
    assert len(tree) == 5
    assert len(tree.values()) == 5
    assert tree.size() == 5
    assert tree.get("jane").tag == "Jane"
    assert tree["jane"].tag == "Jane"
    assert "jane" in tree
    assert "alien" not in tree
    tree.create_node("Alien", "alien", parent="jane")
    assert "alien" in tree
    tree.remove_node("alien")
    assert "alien" not in tree


def test_getitem(tree):
    """Nodes can be accessed via getitem."""
    for node_id in tree:
        assert tree[node_id]

    # assert 'Node access should be possible via getitem.' in str(exc)

    with pytest.raises(NodeNotFound) as exc:
        assert tree['root']

    assert "Node 'root' is not in the tree" in str(exc)


def test_parent(tree):
    for node_id in tree:
        if node_id == tree.root:
            assert tree.parent(node_id) is None
        else:
            assert tree.parent(node_id) in tree.values()


def test_children(tree):
    for node_id in tree:
        children = tree.is_branch(node_id)
        for child in children:
            assert tree[child] in tree.values()
        children = tree.children(node_id)
        for child in children:
            assert child in tree.values()

    with pytest.raises(NodeNotFound) as exc:
        assert tree.is_branch("alien")

    assert "Node 'alien' is not in the tree" in str(exc)


def test_remove_node(tree):
    tree.create_node("Jill", "jill", parent="george")
    tree.create_node("Mark", "mark", parent="jill")
    assert tree.remove_node("jill") == 2
    assert tree.get("jill") is None
    assert tree.get("mark") is None


def test_depth(tree):
    # Try getting the level of this tree
    assert tree.depth() == 2
    tree.create_node("Jill", "jill", parent="george")
    assert tree.depth() == 3
    tree.create_node("Mark", "mark", parent="jill")
    assert tree.depth() == 4

    # Try getting the level of the node
    """
    tree.print()
    Hárry
    |___ Bill
    |    |___ George
    |         |___ Jill
    |              |___ Mark
    |___ Jane
    |    |___ Diane
    """
    assert tree.depth(tree.get("mark")) == 4
    assert tree.depth(tree.get("jill")) == 3
    assert tree.depth(tree.get("george")) == 2
    assert tree.depth("jane") == 1
    assert tree.depth("bill") == 1
    assert tree.depth("hárry") == 0

    # Try getting Exception
    node = Node("Test One", "identifier 1")
    with pytest.raises(NodeNotFound):
        tree.depth(node)


def test_leaves(tree):
    leaves = tree.leaves()
    for node_id in tree.expand_tree():
        assert tree[node_id].is_leaf == (tree[node_id] in leaves)

    leaves = tree.leaves(node_id='jane')
    for node_id in tree.expand_tree(node_id='jane'):
        assert tree[node_id].is_leaf == (tree[node_id] in leaves)


def test_link_past_node(tree):
    tree.create_node("Jill", "jill", parent="hárry")
    tree.create_node("Mark", "mark", parent="jill")
    assert "mark" not in tree.is_branch("hárry")

    tree.link_past_node("jill")
    assert "mark" in tree.is_branch("hárry")


def test_expand_tree(tree):
    # default config
    nodes = [n for n in tree.expand_tree()]
    assert len(nodes) == 5

    # expanding from specific node
    nodes = [n for n in tree.expand_tree(node_id="bill")]
    assert len(nodes) == 2

    # changing into width mode
    nodes = [n for n in tree.expand_tree(mode='width')]
    assert len(nodes) == 5

    # expanding by filters
    nodes = \
        [n for n in tree.expand_tree(filtering=lambda x: x.tag == "Bill")]
    assert len(nodes) == 0
    nodes = \
        [n for n in tree.expand_tree(filtering=lambda x: x.tag != "Bill")]
    assert len(nodes) == 3


def test_move_node(tree):
    diane_parent = tree.parent("diane")
    tree.move_node("diane", "bill")
    assert "diane" in tree.is_branch("bill")
    tree.move_node("diane", diane_parent.id)


def test_paste_tree(tree):
    new_tree = Tree()
    new_tree.create_node("Jill", "jill")
    new_tree.create_node("Mark", "mark", parent="jill")
    tree.paste("jane", new_tree)
    assert "jill" in tree.is_branch("jane")
    tree.remove_node("jill")


def test_rsearch(tree):
    for node_id in ["hárry", "jane", "diane"]:
        assert node_id in tree.rsearch("diane")


def test_subtree(tree):
    subtree_copy = Tree(tree.subtree("jane"), deepcopy=True)
    assert subtree_copy.parent("jane") is None
    subtree_copy["jane"].tag = "Sweetie"
    assert tree["jane"].tag == "Jane"
    assert subtree_copy.level("diane") == 1
    assert subtree_copy.level("jane") == 0
    assert tree.level("jane") == 1


def test_remove_subtree(tree):
    subtree_shallow = tree.remove_subtree("jane")
    assert "jane" not in tree.is_branch("hárry")
    tree.paste("hárry", subtree_shallow)
    assert "jane" in tree.is_branch("hárry")


def test_to_json(tree):
    # assertEqual.__self__.maxDiff = None
    assert tree.to_json() == (
        '{"H\\u00e1rry": {"children": [{"Bill": {"children": ["George"]}}, '
        '{"Jane": {"children": ["Diane"]}}]}}'
    )
    assert tree.to_json(with_data=True) == (
        '{"H\\u00e1rry": {"children": [{"Bill": {"children": '
        '[{"George": {"data": null}}], "data": null}}, '
        '{"Jane": {"children": [{"Diane": {"data": null}}], "data": null}}], '
        '"data": null}}'
    )


def test_siblings(tree):
    assert not tree.siblings("hárry")
    assert tree.siblings("jane")[0].id == "bill"


def test_tree_data(tree):
    class Flower(object):
        def __init__(self, color):
            self.color = color

    tree.create_node("Jill", "jill", parent="jane", data=Flower("white"))
    assert tree["jill"].data.color == "white"
    tree.remove_node("jill")


def test_tree_print_data_property(capsys):
    new_tree = Tree()
    new_tree.print()

    stdout, stderr = capsys.readouterr()
    assert stdout == 'Tree is empty\n'

    class Flower(object):
        def __init__(self, color):
            self.color = color

    new_tree.create_node("Jill", "jill", data=Flower("white"))
    new_tree.print(data_property="color")

    stdout, stderr = capsys.readouterr()
    assert stdout == 'white\n'


def test_level(tree):
    assert tree.level('hárry') == 0
    depth = tree.depth()
    assert tree.level('diane') == depth
    assert tree.level('diane', lambda x: x.id != 'jane') == depth - 1


def test_size(tree):
    assert tree.size(level=2) == 2
    assert tree.size(level=1) == 2
    assert tree.size(level=0) == 1


def test_tree_to_string(tree, tree_as_string):
    assert str(tree) == tree_as_string


def test_tree_print(capsys, tree):
    tree.print()

    stdout, stderr = capsys.readouterr()
    assert stdout == """\
Hárry
├── Jane
│   └── Diane
└── Bill
    └── George
"""


def test_tree_iteration():
    new_tree = Tree()
    assert not new_tree.values()
    nodes = [new_tree.create_node('root_node'),
             new_tree.create_node('second', parent=new_tree.root)]
    for nd in new_tree.values():
        assert nd in nodes


def test_filter_nodes():
    new_tree = Tree()

    assert not tuple(filter(lambda n: True, new_tree.values()))

    nodes = [new_tree.create_node('root_node'),
             new_tree.create_node('second', parent=new_tree.root)]

    assert tuple(filter(lambda n: False, new_tree.values())) == ()
    assert tuple(filter(lambda n: n.is_root, new_tree.values())) == (nodes[0],)
    assert tuple(filter(lambda n: not n.is_root, new_tree.values())) == \
           (nodes[1],)
    assert set(filter(lambda n: True, new_tree.values()))
    assert set(nodes)


def test_loop():
    tree = Tree()
    tree.create_node('a', 'a')
    tree.create_node('b', 'b', parent='a')
    tree.create_node('c', 'c', parent='b')
    tree.create_node('d', 'd', parent='c')

    with pytest.raises(LoopError):
        tree.move_node(source='b', destination='d')


def test_node_to_tree_link():
    tree = Tree()
    node_a = tree.create_node('a', 'a')
    assert node_a.tree is tree
    tree.create_node('b', 'b', parent='a')
    tree.create_node('c', 'c', parent='b')
    tree.create_node('d', 'd', parent='c')
    tree.remove_node(node_a.id)
    assert node_a.tree is None



__author__ = 'Vladimir Bolshakov <vovanbo@gmail.com>'

import json
import copy
from collections import OrderedDict
from typing import Callable, List, MutableMapping, Optional, Union

import ttree.utils
from ttree.common import ASCIIMode, TraversalMode
from ttree.exceptions import (
    NodeNotFound, MultipleRoots, DuplicatedNode, LinkPastRootNode, LoopError
)
from .node import Node


class Tree(OrderedDict):
    """
    The Tree object defines the tree-like structure based on
    :class:`Node` objects. A new tree can be created from scratch without any
    parameter or a shallow/deep copy of another tree. When ``deepcopy=True``,
    a deepcopy operation is performed on feeding ``tree`` parameter and
    *more memory is required to create the tree*.
    """
    def __init__(self, tree: 'Tree' = None, deepcopy: bool = False):
        """Initiate a new tree or copy another tree with a shallow or
        deepcopy copy.
        """
        super(Tree, self).__init__()

        #: id of the root node
        self.root = None

        if tree is not None:
            if not isinstance(tree, Tree):
                raise TypeError('Tree instance is required.')

            self.root = tree.root
            self.__merge_tree(tree, deepcopy)

    def __str__(self) -> str:
        return ttree.utils.print_tree(self, ascii_mode='simple')

    def __getitem__(self, item):
        try:
            return super(Tree, self).__getitem__(item)
        except KeyError:
            raise NodeNotFound(f"Node '{item}' is not in the tree")

    def __setitem__(self, key, value, **kwargs):
        super(Tree, self).__setitem__(key, value, **kwargs)
        if isinstance(value, Node):
            value._tree = self

    def __delitem__(self, key, **kwargs):
        if key in self:
            self[key]._tree = None
        super(Tree, self).__delitem__(key, **kwargs)

    def __merge_tree(self, other: 'Tree', deepcopy: bool = False):
        if deepcopy:
            for node_id in other:
                self[node_id] = copy.deepcopy(other[node_id])
        else:
            self.update(other)

    @property
    def paths_to_leaves(self):
        """
        Use this property to get the identifiers allowing to go from the root
        nodes to each leaf.
        Return a list of list of identifiers, root being not omitted.

        For example:

        .. code-block:: text

            Harry
            |___ Bill
            |___ Jane
            |    |___ Diane
            |         |___ George
            |              |___ Jill
            |         |___ Mary
            |    |___ Mark

        Result:

        .. code-block:: python3

            [['harry', 'jane', 'diane', 'mary'],
             ['harry', 'jane', 'mark'],
             ['harry', 'jane', 'diane', 'george', 'jill'],
             ['harry', 'bill']]
        """
        return [[n for n in self.rsearch(l.id)][::-1] for l in self.leaves()]

    def add_node(self, node: Node, parent: Node = None):
        """
        Add a new node to tree.

        Add a new node object to the tree and make the parent as the root
        by default.
        """
        if not isinstance(node, Node):
            raise TypeError('First parameter must be instance of Node.')

        if node.id in self:
            raise DuplicatedNode(f"Node with ID '{node.id}' "
                                 f"is already exists in tree.")

        pid = parent.id if isinstance(parent, Node) else parent

        if pid is None:
            if self.root is not None:
                raise MultipleRoots('A tree takes one root merely.')

            self.root = node.id
        elif pid not in self:
            raise NodeNotFound(f"Parent node '{pid}' is not in the tree")

        self[node.id] = node
        if pid in self:
            self[pid].add_child(node.id)
        self[node.id].parent = pid

    def children(self, node_id) -> List[Node]:
        """
        Return the children (Node) list of ``node_id``.

        Empty list is returned if ``node_id`` does not exist.
        """
        return [self[i] for i in self.is_branch(node_id)]

    def create_node(self, *args, parent=None, node_cls=Node, **kwargs):
        """
        Create a new node and add it to this tree.

        If ``id`` is absent, a UUID will be generated automatically.
        """
        if not issubclass(node_cls, Node):
            raise ValueError('node_cls must be a subclass of Node.')

        node = node_cls(*args, **kwargs)
        self.add_node(node, parent)
        return node

    def depth(self, node=None) -> int:
        """
        Get the maximum level of this tree or the level of the given node

        .. note ::

            The parameter is node instance rather than node_identifier.

        :param ~ttree.Node node:
        :return: Depth level
        """
        result = 0
        if node is None:
            # Get maximum level of this tree
            for leaf in self.leaves():
                level = self.level(leaf.id)
                result = level if level >= result else result
        else:
            # Get level of the given node
            node_id = node.id if isinstance(node, Node) else node

            if node_id not in self:
                raise NodeNotFound(f"Node '{node_id}' is not in the tree")

            result = self.level(node_id)
        return result

    def expand_tree(self, node_id=None,
                    mode: Union[TraversalMode, str] = TraversalMode.DEPTH,
                    filtering: Callable[[Node], bool] = None,
                    key=None, reverse: bool = False):
        """
        Traverse the tree nodes with different modes.

        ``node_id`` refers to the expanding point to start; ``mode`` refers
        to the search mode (Tree.DEPTH, Tree.WIDTH). ``filter`` refers to the
        function of one variable to act on the :class:`Node` object.
        In this manner, the traversing will not continue to following children
        of node whose condition does not pass the filter. ``key``, ``reverse``
        are present to sort :class:Node objects at the same level.

        Python generator. Loosly based on an algorithm from
        'Essential LISP' by John R. Anderson, Albert T. Corbett, and
        Brian J. Reiser, page 239-241

        UPDATE: the @filtering function is performed on Node object during
        traversing. In this manner, the traversing will not continue to
        following children of node whose condition does not pass the filter.

        UPDATE: the @key and @reverse are present to sort nodes at each
        level.
        """
        node_id = self.root if node_id is None else node_id

        if node_id not in self:
            raise NodeNotFound(f"Node '{node_id}' is not in the tree")

        if filtering is not None and not callable(filtering):
            raise TypeError('Filtering must be callable.')

        mode = mode if isinstance(mode, TraversalMode) else TraversalMode(mode)

        if filtering is not None and not filtering(self[node_id]):
            return

        yield node_id
        queue = (self[i] for i in self[node_id].children)

        if filtering is not None:
            queue = filter(filtering, queue)

        if mode in (TraversalMode.DEPTH, TraversalMode.WIDTH):
            queue = sorted(queue, key=key, reverse=reverse)
            while queue:
                yield queue[0].id
                expansion = sorted(
                    filter(filtering, (self[i] for i in queue[0].children)),
                    key=key, reverse=reverse
                )

                if mode is TraversalMode.DEPTH:
                    queue = expansion + queue[1:]  # depth-first
                elif mode is TraversalMode.WIDTH:
                    queue = queue[1:] + expansion  # width-first

        elif mode is TraversalMode.ZIGZAG:
            # Suggested by Ilya Kuprik (ilya-spy@ynadex.ru).
            stack_fw = []
            queue = list(queue)
            queue.reverse()
            stack = stack_bw = queue
            direction = False
            while stack:
                expansion = filter(filtering,
                                   (self[i] for i in stack[0].children))
                yield stack.pop(0).id
                expansion = list(expansion)
                if direction:
                    expansion.reverse()
                    stack_bw = expansion + stack_bw
                else:
                    stack_fw = expansion + stack_fw
                if not stack:
                    direction = not direction
                    stack = stack_fw if direction else stack_bw

    def is_branch(self, node_id):
        """
        Get the children (only sons) list of the node with ID == node_id.

        Empty list is returned if node_id does not exist
        """
        if node_id is None:
            raise ValueError("First parameter can't be None")

        return self[node_id].children

    def leaves(self, node_id=None):
        """Get leaves from given node."""
        if node_id is None:
            return [n for n in self.values() if n.is_leaf]

        return [self[n] for n in self.expand_tree(node_id) if self[n].is_leaf]

    def level(self, node_id, filtering: Callable[[Node], bool] = None):
        """
        Get the node level in this tree.
        The level is an integer starting with '0' at the root.
        In other words, the root lives at level '0';

        Update: @filtering params is added to calculate level passing
        exclusive nodes.
        """
        return len([n for n in self.rsearch(node_id, filtering)]) - 1

    def link_past_node(self, node_id):
        """
        Remove a node and link its children to its parent.

        Root is not allowed.

        For example, if we have a -> b -> c and delete node b, we are left
        with a -> c
        """
        if self.root == node_id:
            raise LinkPastRootNode('Cannot link past the root node, '
                                   'delete it with remove_node()')

        # Get the parent of the node we are linking past
        parent = self[self[node_id].parent]

        # Set the children of the node to the parent
        for child in self[node_id].children:
            self[child].parent = parent.id

        # Link the children to the parent
        parent.children += self[node_id].children
        # Delete the node
        parent.remove_child(node_id)
        del self[node_id]

    def move_node(self, source, destination):
        """
        Move node (source) from its parent to another parent (destination).
        """
        if source not in self or destination not in self:
            raise NodeNotFound

        if self.is_ancestor(source, destination):
            raise LoopError

        parent = self[source].parent
        self[parent].remove_child(source)
        self[destination].add_child(source)
        self[source].parent = destination

    def is_ancestor(self, ancestor, grandchild) -> bool:
        parent = self[grandchild].parent
        child = grandchild

        while parent is not None:
            if parent == ancestor:
                return True

            child = self[child].parent
            parent = self[child].parent

        return False

    def parent(self, node_id) -> Optional[Node]:
        """
        Obtain specific node's parent (Node instance).

        Return None if the parent is None or does not exist in the tree.
        """
        pid = self[node_id].parent
        if pid is None or pid not in self:
            return None

        return self[pid]

    def paste(self, node_id, new_tree: 'Tree', deepcopy: bool = False):
        """
        Paste a new tree to an existing tree, with ``node_id`` becoming the
        parent of the root of this new tree.

        Update: add @deepcopy of pasted tree.
        """
        if not isinstance(new_tree, Tree):
            raise TypeError('Instance of Tree is required as '
                            '"new_tree" parameter.')

        if node_id is None:
            raise ValueError('First parameter can not be None')

        if node_id not in self:
            raise NodeNotFound(f"Node '{node_id}' is not in the tree")

        set_joint = set(new_tree) & set(self)  # joint keys
        if set_joint:
            # TODO: a deprecated routine is needed to avoid exception
            raise ValueError(f'Duplicated nodes {list(set_joint)} exists.')

        self.__merge_tree(new_tree, deepcopy)

        self[node_id].add_child(new_tree.root)
        self[new_tree.root].parent = node_id

    def remove_node(self, node_id) -> int:
        """
        Remove a node indicated by 'id'; all the successors are
        removed as well.

        Return the number of removed nodes.
        """
        if node_id is None:
            return 0

        parent = self[node_id].parent

        removed = [n for n in self.expand_tree(node_id)]
        for id_ in removed:
            del self[id_]

        # Update its parent info
        if parent:
            self[parent].remove_child(node_id)

        return len(removed)

    def remove_subtree(self, node_id) -> 'Tree':
        """
        Return a subtree deleted from this tree. If node_id is None, an
        empty tree is returned.
        For the original tree, this method is similar to
        `remove_node(self,node_id)`, because given node and its children
        are removed from the original tree in both methods.
        For the returned value and performance, these two methods are
        different:

            `remove_node` returns the number of deleted nodes;
            `remove_subtree` returns a subtree of deleted nodes;

        You are always suggested to use `remove_node` if your only to
        delete nodes from a tree, as the other one need memory
        allocation to store the new tree.
        """
        subtree = Tree()
        if node_id is None:
            return subtree

        subtree.root = node_id
        parent = self[node_id].parent
        self[node_id].parent = None  # reset root parent for the new tree

        removed = [n for n in self.expand_tree(node_id)]
        for id_ in removed:
            subtree[id_] = self.pop(id_)

        # Update its parent info
        self[parent].remove_child(node_id)
        return subtree

    def rsearch(self, node_id, filtering: Callable[[Node], bool] = None):
        """
        Search the tree from ``node_id`` to the root along links reservedly.

        Parameter ``filter`` refers to the function of one variable to act
        on the :class:`Node` object.
        """
        if node_id is None:
            return

        if node_id not in self:
            raise NodeNotFound(f"Node '{node_id}' is not in the tree")

        if filtering is not None and not callable(filtering):
            raise TypeError('Filtering must be a callable.')

        current = node_id
        while current is not None:
            if filtering is None or filtering(self[current]):
                yield current

            # subtree() hasn't update the parent
            current = self[current].parent if self.root != current else None

    def save2file(self, filename, node_id=None, id_hidden=True,
                  filtering=None, key=None, reverse=False,
                  ascii_mode=ASCIIMode.ex, data_property=None):
        """
        Save the tree into file for offline analysis.

        :param filename: Export file name
        :param node_id: Traversal root node ID
        :param id_hidden: Is ID hidden?
        :param filtering: Filtering callable
        :param key: Sorting key callable
        :param reverse: Reverse mode?
        :param ascii_mode: ASCII mode
        :param data_property: Data property name
        """
        with open(filename, 'ab') as fp:
            ttree.utils.print_tree(
                self, node_id, id_hidden, filtering, key, reverse,
                ascii_mode, data_property, func=lambda n: fp.write(n + b'\n')
            )

    def print(self, node_id=None, id_hidden=True, filtering=None,
              key=None, reverse=False, ascii_mode=ASCIIMode.ex,
              data_property=None):
        """
        Print the tree structure in hierarchy style.

        You have three ways to output your tree data, i.e., stdout with
        ``print()``, plain text file with ``save2file()``, and json string
        with ``to_json()``. The former two use the same backend to generate
        a string of tree structure in a text graph.

        :param node_id: refers to the expanding point to start
        :param id_hidden: refers to hiding the node ID when printing
        :param filtering: refers to the function of one variable to act on the
            :class:`Node` object. In this manner, the traversing will not
            continue to following children of node whose condition does
            not pass the filter.
        :param key: are present to sort :class:`Node` object in the same level
        :param reverse: reverse mode
        :param ascii_mode: ASCII mode
        :param data_property: refers to the property on the node data object
            to be printed.
        """
        try:
            ttree.utils.print_tree(
                self, node_id, id_hidden, filtering,
                key, reverse, ascii_mode, data_property,
                func=print
            )
        except NodeNotFound:
            print('Tree is empty')

    def siblings(self, node_id) -> List[Node]:
        """
        Return the siblings of given ``node_id``.

        If ``node_id`` is root or there are no siblings, an empty list
        is returned.
        """
        siblings = []

        if node_id != self.root:
            pid = self[node_id].parent
            siblings = [self[c] for c in self[pid].children if c != node_id]

        return siblings

    def size(self, level: int = None) -> int:
        """
        Get the number of nodes of the whole tree if @level is not
        given. Otherwise, the total number of nodes at specific level
        is returned.

        @param level The level number in the tree. It must be between
        [0, tree.depth].

        Otherwise, InvalidLevelNumber exception will be raised.
        """
        if level is None:
            return len(self)

        if not isinstance(level, int):
            raise TypeError(f"Level should be an integer instead "
                            f"of '{type(level)}'")

        return len(
            [node for node in self.values()
             if self.level(node.id) == level]
        )

    def subtree(self, node_id) -> 'Tree':
        """
        Return a shallow COPY of subtree with node_id being the new root.
        If node_id is None, return an empty tree.
        If you are looking for a deepcopy, please create a new tree
        with this shallow copy,

        e.g.
            new_tree = Tree(t.subtree(t.root), deep=True)

        This line creates a deep copy of the entire tree.
        """
        result = self.__class__()
        if node_id is None:
            return result

        if node_id not in self:
            raise NodeNotFound(f"Node '{node_id}' is not in the tree")

        result.root = node_id
        for subtree_node in self.expand_tree(node_id):
            result[self[subtree_node].id] = self[subtree_node]

        return result

    def to_dict(self, node_id=None, key=None, sort=True, reverse=False,
                with_data=False) -> MutableMapping:
        """transform self into a dict"""

        node_id = self.root if node_id is None else node_id
        node_tag = self[node_id].tag
        result = {node_tag: {'children': []}}
        if with_data:
            result[node_tag]['data'] = self[node_id].data

        if self[node_id].expanded:
            queue = (self[i] for i in self[node_id].children)
            if sort:
                sort_options = {'reverse': reverse}
                if key is not None:
                    sort_options['key'] = key
                queue = sorted(queue, **sort_options)

            result[node_tag]['children'] = [
                self.to_dict(n.id, with_data=with_data,
                             sort=sort, reverse=reverse)
                for n in queue
            ]

            if not result[node_tag]['children']:
                result = (
                    self[node_id].tag
                    if not with_data
                    else {node_tag: {'data': self[node_id].data}}
                )

        return result

    def to_json(self, with_data=False, sort=True, reverse=False) -> str:
        """Return the json string corresponding to self"""
        return json.dumps(
            self.to_dict(with_data=with_data, sort=sort, reverse=reverse)
        )



#!/usr/bin/env python
import uuid

from collections import Sequence, MutableMapping, Set


class Node:
    """
    Represent a tree node.

    A :class:`Node` object contains basic properties such as node id,
    node tag, parent node, children nodes etc., and some operations for a node.

    Nodes are elementary objects which are stored in a dictionary
    of a Tree. Use `data` attribute to store node-specific data.
    """
    def __init__(self, tag=None, id=None, expanded=True, data=None, tree=None):
        """Create a new Node object to be placed inside a Tree object"""

        #: if given as a parameter, must be unique
        self._id = None
        self._set_id(id)

        #: None or something else
        #: if None, self._id will be set to the id's value.
        self._tag = self._id if tag is None else tag

        #: boolean
        self.expanded = expanded

        #: id of the parent's node
        self._parent = None
        #: id(s) of the children node(s)
        self._children = list()

        self._tree = tree
        #: User payload associated with this node.
        self.data = data

    def __repr__(self):
        kwargs = (f"tag={self.tag!r}",
                  f"id={self.id!r}",
                  f"data={self.data!r}")
        return f"{self.__class__.__name__}({', '.join(kwargs)})"

    def __lt__(self, other):
        return self.tag < other.tag

    def _set_id(self, node_id):
        """Initialize self._set_id"""
        self._id = uuid.uuid1() if node_id is None else node_id

    @property
    def id(self):
        """
        Return the value of `_id`.

        The unique ID of a node within the scope of a tree. This attribute
        can be accessed and modified with ``.`` and ``=`` operator
        respectively.
        """
        return self._id

    @id.setter
    def id(self, value):
        """Set the value of `_id`."""
        if value is None:
            raise ValueError("Node ID can not be None")

        self._set_id(value)

    @property
    def tag(self):
        """
        Return the value of `_tag`.

        The readable node name for human. This attribute can be accessed and
        modified with ``.`` and ``=`` operator respectively.
        """
        return self._tag

    @tag.setter
    def tag(self, value):
        """Set the value of `_tag`."""
        self._tag = value

    @property
    def parent(self):
        """
        Return the value of `_parent`.

        The parent ID of a node. This attribute can be accessed and modified
        with ``.`` and ``=`` operator respectively.
        """
        return self._parent

    @parent.setter
    def parent(self, node_id):
        """Set the value of `_parent`."""
        self._parent = node_id

    @property
    def children(self):
        """
        Return the value of `_children`.

        With a getting operator, a list of IDs of node's children is obtained.
        With a setting operator, the value can be list, set, or dict.
        For list or set, it is converted to a list type by the package;
        for dict, the keys are treated as the node IDs.
        """
        return self._children

    @children.setter
    def children(self, value):
        """Set the value of `_children`."""
        if value is None:
            self._children = list()
        elif isinstance(value, (Sequence, Set)):
            self._children = list(value)
        elif isinstance(value, MutableMapping):
            self._children = list(value.keys())
        else:
            raise ValueError('Sequence, Set or MutableMapping '
                             'are allowed values for children only.')

    @property
    def is_leaf(self):
        """
        Check if the node has children.

        Return False if the ``children`` is empty or None.
        """
        return not self.children

    @property
    def is_root(self):
        """Check if the node is the root of present tree."""
        return self._parent is None

    @property
    def tree(self):
        """Return tree instance of node."""
        return self._tree

    def add_child(self, node_id):
        """Add child (indicated by the ``node_id`` parameter) of a node."""
        if node_id is not None:
            self._children.append(node_id)

    def remove_child(self, node_id):
        """Remove child (indicated by the ``node_id`` parameter) of a node."""
        if node_id is not None and node_id in self._children:
            self._children.remove(node_id)



from setuptools import setup
from ttree import __version__

setup(
    name="ttree",
    version=__version__,
    url='https://github.com/vovanbo/ttree',
    author='Vladimir Bolshakov',
    author_email='vovanbo@gmail.com',
    description='Taxonomy tree data structure implementation.',
    long_description='This is a taxonomy tree data structure implementation '
                     'in Python.',
    license="Apache License, Version 2.0",
    packages=['ttree'],
    keywords=['data structure', 'tree', 'tools', 'taxonomy'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: Freely Distributable',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
